"""Eargait: The Gait and Movement Analysis Package unsing Ear-Worn IMU sensors."""
from eargait.spatial_params.spatial_params_example_class import SpatialParamsExample

__all__ = ["SpatialParamsExample"]
